<?php
/**
 * Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2019 All Rights Reserved
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 */


namespace App\Base\Controllers;

use App\Base\Base;
use App\Base\Controller;
use App\Base\Exception;
use App\Base\Response;


class Spa extends Controller
{
    /**
     * Path to "cloudlinux-cli.py" script.
     *
     * @var string
     */
    public $cloudlinuxCLI = '/usr/bin/sudo /usr/share/l.v.e-manager/utils/cloudlinux-cli.py';
    public $cloudlinuxUserCLI = '/usr/share/l.v.e-manager/utils/cloudlinux-cli-user.py';

    /**
     * the type of user to run script.
     *
     * @var string
     */
    protected $owner = 'user';

    /**
     * Login of user.
     *
     * @var
     */
    protected $userLogin;

    /**
     * Current domain for user.
     *
     * @var
     */
    protected $userDomain;

    /**
     * ID of user.
     *
     * @var
     */
    protected $userUID;

    /**
     * Required parameters.
     *
     * @var array
     */
    public $userInfo;


    public function init()
    {
        $this->owner = Base::load()->getOwner();
        if (in_array($this->owner, array(Base::OWNER_USER, Base::OWNER_RESELLER))) {
            $this->prepareUserInfo();
        }
    }

    protected function prepareUserInfo()
    {
        $this->userInfo = array(
            'username' => $this->getLogin(),
            'lve-id' => $this->getUID(),
        );
    }


    protected function getLogin()
    {
        if (is_null($this->userLogin)) {
            $this->userLogin = Base::load()->billingManager->getLogin();
        }
        return $this->userLogin;
    }

    protected function getUserDomain()
    {
        if (is_null($this->userDomain)) {
            $this->userDomain = Base::load()->billingManager->getCurrentDomain();
        }
        return $this->userDomain;
    }

    protected function getUID()
    {
        if (is_null($this->userUID)) {
            $this->userUID = Base::load()->billingManager->getUID();
        }
        return $this->userUID;
    }

    /**
     * Processes of incoming post request.
     */
    public function processRequest()
    {
        $this->_checkVulnerabilities();

        if (!Base::load()->hasParam('command')) {
            throw new Exception('COMMAND NOT SPECIFIED');
        }

        $data['owner'] = $this->owner;
        $data['command'] = Base::load()->getPost('command');

        if (defined("PLUGIN_NAME")) {
            $data['plugin_name'] = constant("PLUGIN_NAME");
        }

        if (Base::load()->hasParam('method')) {
            $data['method'] = Base::load()->getPost('method');
        }

        if (Base::load()->hasParam('params')) {
            $data['params'] = Base::load()->getPost('params');
        }

        if (!is_null($_POST['attachments'])) {
            $fileList = array();
            // check for uploaded files
            $files = $_POST['attachments'];
            foreach ($files as $fileNum => $tempName) {
                if (!empty($tempName) && strpos($tempName, '/home/tmp/') === 0) {
                    $fileList[] = array(
                        'file' => $tempName,
                        'name' => basename(substr($tempName, 0, -6))
                    );
                }
            }
            if (!empty($fileList)) {
                $data['attachments'] = $fileList;
            }
        }

        if (in_array($this->owner, array(Base::OWNER_USER, Base::OWNER_RESELLER))) {
            $data['user_info'] = $this->userInfo;
        }

        if (Base::load()->hasParam('mockJson')) {
            $data['mockJson'] = Base::load()->getPost('mockJson');
        }

        if (Base::load()->hasParam('lang')) {
            $data['lang'] = Base::load()->getPost('lang');
        }

        $fullCommandStr = sprintf(
            "%s --data=%s 2>&1",
            $this->owner == Base::OWNER_USER ? $this->cloudlinuxUserCLI : $this->cloudlinuxCLI,
            base64_encode(json_encode($data))
        );

        putenv('LC_ALL=en_US.UTF-8');

        ob_start();
        passthru($fullCommandStr);
        $responseInJson = ob_get_contents();
        ob_end_clean();


        if ($this->owner == Base::OWNER_USER && !file_exists($this->cloudlinuxUserCLI)) {
            switch (constant("PLUGIN_NAME")) {
                case 'nodejs_selector':
                    $pluginTitle = 'Node.js Selector';
                    break;
                case 'python_selector':
                    $pluginTitle = 'Python Selector';
                    break;
                default:
                    $pluginTitle = 'Plugin';
            }
            throw new Exception('',
                array(
                    'result' => '',
                    'code' => 503,
                    'error_id' => 'ERROR.not_available_plugin',
                    'context'  => array(
                        'pluginName' => $pluginTitle
                    ),
                    'icon' => 'disabled'
                )
            );
        }

        $response = json_decode($responseInJson);
        if (is_null($response) && !empty($responseInJson)) {
            throw new Exception('ERROR.wrong_received_data', array(
                'details' => $responseInJson
            ));
        }

        if (isset($response->result) && $response->result === 'file') {
            Response::serveFile(
                $response->filepath, $response->filesize, $data,
                $this->owner == Base::OWNER_USER ? $this->cloudlinuxUserCLI : $this->cloudlinuxCLI
            );
            return;
        } else
            if (isset($response->result) && $response->result !== 'success' && $response->result !== 'rollback') {
                throw new Exception($response->result, $response);
            } else
                if (empty($response)) {
                    throw new Exception('RESPONSE OF COMMAND IS EMPTY');
                }


        echo Response::asJson($response);
    }

    private function _checkVulnerabilities()
    {
        $this->_checkCSRFToken();
        //$this->_checkReferer(); // DA does't pass HTTP_REFERER header
    }

    private function _checkCSRFToken()
    {
        if (!isset($_COOKIE['csrftoken'])
            || $_COOKIE['csrftoken'] !== $_POST['csrftoken']
        ) {
            throw new Exception(
                'BAD FORGERY PROTECTION TOKEN',
                array('logoutSignal' => true)
            );
        }
    }
}
